ENDPOINTS = {
    "project_create": "/api/project/create",
    "project_create_from_github": "/api/v1/project/create_from_github",
    "project_list": "/api/project/list",
    "project_sessions": "/api/project/sessions",
    "session_autosync_add_device": "/api/session/autosync_add_device",
    "session_autosync_started": "/api/session/autosync_started",
    "session_commands": "/api/session/commands",
    "session_create": "/api/session/create",
    "session_describe": "/api/session/describe",
    "session_execute": "/api/v1/session/execute",
    "session_execution_logs": "/api/session/execution_logs",
    "session_monitor_logs": "/api/session/monitor_logs",
    "session_finish_command": "/api/session/finish_command",
    "session_kill_command": "/api/session/kill_command",
    "session_list": "/api/session/list",
    "session_overview": "/api/session/overview",
    "session_report_command": "/api/session/report_command",
    "session_startup_log": "/api/v1/session/startup_log",
    "session_snapshot": "/api/v1/session/snapshot",
    "session_start": "/api/session/start",
    "session_stop": "/api/session/stop",
    "session_sync": "/api/session/sync",
    "session_upload_command_logs": "/api/session/upload_command_logs",
    "snapshot_archive": "/api/snapshot/archive",
    "snapshot_create": "/api/snapshot/create",
    "snapshot_delete": "/api/snapshot/delete",
    "snapshot_describe": "/api/v1/snapshot/describe",
    "snapshot_get": "/api/snapshot/get",
    "snapshot_list": "/api/snapshot/list",
    "user_info": "/api/user/info",
    "user_login": "/api/user/login",
    "user_logout": "/api/user/logout",
    "user_register": "/api/user/register",
    "user_server_session_token": "/api/user/server_session_token",
    "user_get_temporary_aws_credentials": "/api/user/temporary_credentials",
    "user_aws_get_anyscale_account": "/api/v1/user/aws/anyscale_account",
    "user_setup_aws_cross_account_role": "/api/v1/user/aws/cross_account_role",
}
